/*
btinit.c - Release 0.1.0  Btree initialize database
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

#include <bthash.h>

#define BLKHDR        8
#define RCDHDR        4
#define MAXBLKSZ  65536
#define MINRPN        3
#define MAXRPN     1023
#define MAXRCDLEN 32000

int BTGetBlkInit(hndl,blklen,blk)
int hndl;
int blklen;
unsigned char *blk;
   {
   int rdlen;
   rdlen = read(hndl,blk,blklen);
   if (rdlen == blklen)
      return(0);
   else if (rdlen < 0)
      {
      fprintf(stderr,"BTGetBlkInit: errno %d\n", errno);
      perror("Read error");
      return(SYS_ERR + errno);
      } /* read error */
   else
      {
      fprintf(stderr,"BTGetBlkInit: read error "
	 "(rdlen %d)\n", rdlen);
      return(BTREE_ERR + 107);
      } /* read error */
   } /* BTGetBlkInit */

int BTPutBlkInit(hndl,blklen,blk)
int hndl;
int blklen;
unsigned char *blk;
   {
   int wrtlen;
   wrtlen = write(hndl,blk,blklen);
   if (wrtlen == blklen)
      return(0);
   else if (wrtlen < 0)
      {
      fprintf(stderr,"BTPutBlkInit: errno %d\n", errno);
      perror("Write error");
      return(SYS_ERR + errno);
      } /* write error */
   else
      {
      fprintf(stderr,"BTPutBlkInit: write error "
	 "(wrtlen %d)", wrtlen);
      return(BTREE_ERR + 108);
      } /* write error */
   return(0);
   } /* BTPutBlkInit */

int BTClsInit(hndl)
int hndl;
   {
   int rslt;
   rslt = close(hndl);
   if (rslt < 0)
      {
      perror("BTClsInit: close error");
      return(SYS_ERR + errno);
      } /* if close error */
   return(0);
   } /* BTClsInit */

int BTBldHdrInit(hndl,blksz,nodesz,
   rcdlen,keylen,keyofst,dletofst)
int hndl;
int blksz;
int nodesz;
int rcdlen;
int keylen;
int keyofst;
int dletofst;
   {
   int rslt;
   int hdrsz;
   btreefmt *blk;
   unsigned char *p,*q;
   hdrsz = blksz + ROOT;
   blk = (btreefmt *) malloc(hdrsz+16);
   if (blk == NULL)
      {
      fprintf(stderr,"BTBldHdrInit: malloc error\n");
      return(BTREE_ERR + 110);
      } /* malloc err */
   p = (unsigned char *) blk;
   q = (unsigned char *) blk + hdrsz;
   while (p < q) *p++ = '\0';
   strcpy(blk->signature,BT_SIGNATURE);
   blk->blksz    = blksz;
   blk->nodesz   = nodesz;   /* records per node */
   blk->rcdlen   = rcdlen;
   blk->keylen   = keylen;
   blk->keyofst  = keyofst;
   blk->dletofst = dletofst;
   blk->blkhdr   = BLKHDR;
   blk->rcdhdr   = RCDHDR;
   blk->rootofst = ROOT;
   rslt = BTPutBlkInit(hndl,hdrsz,blk);
   free(blk);
   return(rslt);
   } /* BTBldHdrInit */

int BTOpnInitOld(dbname,hndl)
char *dbname;
int *hndl;
   {
   int rslt;
   int oldhndl;
   unsigned char buf[32];
   *hndl = -1;
   oldhndl = open(dbname,O_RDONLY);
   if (oldhndl < 0)
      {
      fprintf(stderr,"BTOpnInitOld: %s: "
	 "open error\n",
         dbname);
      perror("Open error");
      return(SYS_ERR + errno);
      } /* open err */
   rslt = lockf(oldhndl,F_TEST,0);
   if (rslt < 0)
      {
      fprintf(stderr,"BTOpnInitOld: lock test "
         "error on %s\n", dbname);
      perror("Lock test error");
      BTClsInit(oldhndl);
      return(SYS_ERR + errno);
      } /* if lock test failed */
   rslt = BTGetBlkInit(oldhndl,16,buf);
   if (rslt)
      {
      BTClsInit(oldhndl);
      return(rslt);
      } /* BTGetBlkInit err */
   rslt = strncmp(buf,BT_SIGNATURE,16); 
   if (rslt)
      {
      fprintf(stderr,"BTOpnInitOld: filename %s "
         "for non-btree file\n",
         dbname);
      BTClsInit(oldhndl);
      return(BTREE_ERR + 106);
      } /* not btree file */
   else
      {
      rslt = BTClsInit(oldhndl);
      if (rslt) return(rslt);
      } /* close after good signature */
   *hndl = open(dbname,O_RDWR|O_TRUNC);
   if (*hndl < 0)
      {
      fprintf(stderr,"BTOpnInitOld: %s: "
	 "open read/write\n",
         dbname);
      perror("Open error");
      return(SYS_ERR + errno);
      } /* if open error */
   rslt = lockf(*hndl,F_TLOCK,0);
   if (rslt < 0)
      {
      fprintf(stderr,"BTOpnInitOld: lock "
         "error on %s\n", dbname);
      perror("Lock error");
      BTClsInit(*hndl);
      return(SYS_ERR + errno);
      } /* if lock failed */
   return(0);
   } /* BTOpnInitOld */

int BTOpnInitNew(dbname,hndl)
char *dbname;
int *hndl;
   {
   int rslt;
   int oldhndl;
   oldhndl = open(dbname,O_RDONLY);
   if (oldhndl > 0)
      {
      fprintf(stderr,"BTOpnInitNew: %s: "
	 "file exists\n",
         dbname);
      BTClsInit(oldhndl);
      return(BTREE_ERR + 109);
      } /* open err */
   *hndl = open(dbname,
      O_RDWR|O_CREAT|O_EXCL,
      S_IRUSR | S_IWUSR);
   if (*hndl < 0)
      {
      fprintf(stderr,"BTOpnInitNew: %s: "
	 "open read/write\n",
         dbname);
      perror("Open error");
      return(SYS_ERR + errno);
      }
   rslt = lockf(*hndl,F_TLOCK,0);
   if (rslt < 0)
      {
      fprintf(stderr,"BTOpnInitNew: lock "
         "error on %s\n", dbname);
      perror("Lock error");
      BTClsInit(*hndl);
      return(SYS_ERR + errno);
      } /* if lock failed */
   return(0);
   } /* BTOpnInitNew */

int BTOpnInitAll(dbname,hndl)
char *dbname;
int *hndl;
   {
   int rslt;
   int oldhndl;
   unsigned char buf[32];
   *hndl = -1;
   oldhndl = open(dbname,O_RDONLY);
   if (oldhndl < 0)
      {
      if (errno == ENOENT);
      else
	 {
         fprintf(stderr,"BTOpnInitAll: %s: "
	    "open error\n",
            dbname);
         perror("Open error");
         return(SYS_ERR + errno);
	 } /* serious open err */
      } /* open err */
   else if (oldhndl > 0)
      {
      rslt = lockf(oldhndl,F_TEST,0);
      if (rslt < 0)
         {
         fprintf(stderr,"BTOpnInitAll: lock test "
            "error on %s\n", dbname);
         perror("Lock test error");
         BTClsInit(oldhndl);
         return(SYS_ERR + errno);
         } /* if lock test failed */
      rslt = BTGetBlkInit(oldhndl,16,buf);
      if (rslt) return(rslt);
      rslt = strncmp(buf,BT_SIGNATURE,16); 
      if (rslt)
	 {
	 fprintf(stderr,"BTOpnInitAll: filename %s "
	    "for non-btree file\n",
	    dbname);
	 BTClsInit(oldhndl);
	 return(BTREE_ERR + 115);
	 } /* not btree file */
      else
	 {
	 rslt = BTClsInit(oldhndl);
	 if (rslt) return(rslt);
	 } /* close after good signature */
      } /* if good open rdonly */
   *hndl = open(dbname,O_RDWR|O_TRUNC);
   if (*hndl < 0)
      {
      if (errno == ENOENT)
	 {
	 fprintf(stderr,"Creating %s\n", dbname);
         *hndl = open(dbname,
	    O_RDWR|O_CREAT|O_EXCL,
            S_IRUSR | S_IWUSR);
         if (*hndl < 0)
            {
	    fprintf(stderr,"BTOpnInitAll: %s: "
	       "open read/write\n",
	       dbname);
	    perror("Open error");
            return(SYS_ERR + errno);
            }
	 } /* new database */
      else
	 {
	 fprintf(stderr,"BTOpnInitAll: %s: "
	    "open read/write\n",
	    dbname);
	 perror("Open error");
         return(SYS_ERR + errno);
	 } /* genuine open error */
      } /* if open error */
   rslt = lockf(*hndl,F_TLOCK,0);
   if (rslt < 0)
      {
      fprintf(stderr,"BTOpnInitAll: lock "
         "error on %s\n", dbname);
      perror("Lock error");
      BTClsInit(*hndl);
      return(SYS_ERR + errno);
      } /* if lock failed */
   return(0);
   } /* BTOpnInitAll */

int BTOpnInit(dbname,truncflag,
   blksz,nodesz,rcdlen,
   keylen,keyofst,dletofst)
char *dbname;
int blksz;
int nodesz;
int rcdlen;
int keylen;
int keyofst;
   {
   int rslt;
   int hndl;
   if (truncflag == INITOLD)
      rslt = BTOpnInitOld(dbname,&hndl);
   else if (truncflag == INITNEW)
      rslt = BTOpnInitNew(dbname,&hndl);
   else if (truncflag == INITALL)
      rslt = BTOpnInitAll(dbname,&hndl);
   if (rslt) return(rslt);

   rslt = BTBldHdrInit(hndl,blksz,nodesz,
      rcdlen,keylen,keyofst,dletofst);
   if (rslt)
      {
      BTClsInit(hndl);
      return(rslt);
      } /* if bad bldhdr */
   return(BTClsInit(hndl));
   } /* BTOpnInit */

int BTValidateInit(dbname,truncflag,
   nodesz,rcdlen,keylen,keyofst,dletofst)
char *dbname;
int truncflag;
int nodesz;
int rcdlen;
int keylen;
int keyofst;
   {
   int blksz;
   blksz = ((rcdlen + RCDHDR) * nodesz) + BLKHDR;
   if (blksz > MAXBLKSZ)
      {
      fprintf(stderr,"btinit: block size %d "
	 "exceeds maximum %d\n",
	 blksz, MAXBLKSZ);
      fprintf(stderr,"btinit: Adjust "
	 "records-per-node "
	 "and record-length\n");
      return(BTREE_ERR + 111);
      } /* blksz */
   if (keylen > rcdlen)
      {
      fprintf(stderr,"btinit: key-length %d "
         "exceeds record-length %d\n",
	 keylen, rcdlen);
      return(BTREE_ERR + 112);
      }
   if (keyofst > rcdlen - keylen)
      {
      fprintf(stderr,"btinit: key-offset %d "
         "exceeds the record-length %d\n",
	 keyofst, rcdlen);
      fprintf(stderr,"    minus the key-length %d\n",
	 keylen);
      return(BTREE_ERR + 113);
      }
   if (dletofst >= keyofst
      && dletofst < keyofst + keylen)
      {
      fprintf(stderr,"btinit: delete-offset %d "
         "occupies key area %d to %d\n",
	 dletofst, keyofst, keyofst + keylen - 1);
      return(BTREE_ERR + 117);
      }
   return(BTOpnInit(dbname,truncflag,
      blksz,nodesz,rcdlen,
      keylen,keyofst,dletofst));
   } /* BTValidateInit */

int btinit(dbname,truncflag,nodesz,
   rcdlen,keylen,keyofst,dletofst)
char *dbname;
int truncflag;
int nodesz;
int rcdlen;
int keylen;
int keyofst;
int dletofst;
   {
   if (truncflag == INITNEW
      || truncflag == INITOLD
      || truncflag == INITALL);
   else
      {
      fprintf(stderr,"btinit: invalid "
	 "truncate flag %d\n",
	 truncflag);
      return(BTREE_ERR + 114);
      } /* bad truncflag */
   if (nodesz < MINRPN || nodesz > MAXRPN)
      {
      fprintf(stderr,"btinit: invalid "
         "records-per-node %d\n",
         nodesz);
      return(BTREE_ERR + 101);
      }
   if (!(nodesz & 1))
      {
      fprintf(stderr,"btinit: records-per-node "
         "%d must be odd\n",
	 nodesz);
      return(BTREE_ERR + 102);
      }
   if (rcdlen < 1 || rcdlen > MAXRCDLEN)
      {
      fprintf(stderr,"btinit: invalid "
         "record-length %d\n",
         rcdlen);
      return(BTREE_ERR + 103);
      }
   if (keylen < 1 || keylen > MAXRCDLEN)
      {
      fprintf(stderr,"btinit: invalid "
         "key-length %d\n",
         keylen);
      return(BTREE_ERR + 104);
      }
   if (keyofst < 0 || keyofst > MAXRCDLEN-1)
      {
      fprintf(stderr,"btinit: invalid "
         "key-offset %d\n",
         keyofst);
      return(BTREE_ERR + 105);
      }
   if (dletofst < 0 || dletofst > MAXRCDLEN-1)
      {
      fprintf(stderr,"btinit: invalid "
         "delete-offset %d\n",
         dletofst);
      return(BTREE_ERR + 116);
      }
   return(BTValidateInit(dbname,truncflag,
      nodesz,rcdlen,keylen,
      keyofst,dletofst));
   } /* btinit */
