/*
bldhsh.c Version 0.1.0  Hash build test database
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* Usage: bldhsh filename */

#include <bthash.h>

#define MAXLMT 999999

void putstx(pgm)
char *pgm;
   {
   fprintf(stderr,"Usage:   %s filename "
      "#records\n", pgm);
   fprintf(stderr,"Example: %s test.hsh 100\n", pgm);
   fprintf(stderr,"#records is 1 - %d\n",
      MAXLMT);
   exit(1);
   } /* putstx */

int main(argc,argv)
int argc;
char **argv;
   {
   int opnerr;
   int i;
   int rslt;
   int lmt;
   int rcdtot;
   int ovfltot;
   int emptytot;
   int freeblk;
   int totdups;
   int totnf;
   int updtnf;
   hashfmt *hshhdr;
   unsigned char *rndSeed;
   unsigned char *p,*q;
   unsigned char str[4096];
   unsigned char buf[4096*8];
   unsigned char dbname[256];
   if (argc != 3) putstx(*argv);
   strcpy(dbname,*(argv+1));
   lmt = (int) atoi(*(argv+2));
   if (lmt < 1 || lmt > MAXLMT) putstx(*argv);

   rndSeed = (unsigned char *) rndinit();
   if (rndSeed == NULL)
      {
      fprintf(stderr,"bldhsh: rndinit error\n");
      exit(1);
      } /* rndinit err */

   hshhdr = (hashfmt *) hopen(dbname,UPDT,LOCK,&opnerr);
   if (hshhdr == NULL)
      {
      fprintf(stderr,"bldhsh: open "
	 "error %d\n", opnerr);
      free(rndSeed);
      return(0);
      } /* open error */
   /* process starts here */

   rslt = 0;
   totdups = totnf = updtnf = 0;
   i = lmt;
   while (i--)
      {
      p = (char *) str;
      q = (char *) p + hshhdr->keyofst;
      while (p < q)
	 {
	 *p++ = rndnum(26,rndSeed) + 'a';
	 } /* for each char in rcd */
      q = (char *) p + hshhdr->keylen;
      while (p < q)
	 {
	 *p++ = rndnum(26,rndSeed) + 'A';
	 } /* for each char in rcd */
      q = (char *) str + hshhdr->rcdlen;
      while (p < q)
	 {
	 *p++ = rndnum(26,rndSeed) + 'a';
	 } /* for each char in rcd */
      *p = '\0';
      rslt = hisrt(hshhdr,str);
      if (rslt == 1)
	 {
	 totdups++;
	 } /* dup key */
      else if (rslt)
	 {
	 fprintf(stderr,"bldhsh: "
	    "hisrt error %d\n",
	    rslt);
	 exit(1);
	 } /* hisrt err */
      rslt = hget(hshhdr,
	 str+hshhdr->keyofst,buf);
      if (rslt == 1)
	 {
	 totnf++;
	 } /* total not found */
      else if (rslt)
	 {
	 fprintf(stderr,"bldhsh: hget "
	    "found %d %s %s\n",
	    rslt, str, buf);
	 exit(1);
	 } /* found */
      p = (unsigned char *) str;
      q = (unsigned char *) p + hshhdr->keyofst;
      while (p < q) *p++ = 'z';
      rslt = hupdt(hshhdr,str);
      if (rslt == 1)
	 {
	 updtnf++;
	 } /* total update not found */
      else if (rslt)
	 {
	 fprintf(stderr,"bldhsh: hupdt "
	    "not found %d %s %s\n",
	    rslt, str, buf);
	 exit(1);
	 } /* found */
      } /* insert loop */

   hsize(hshhdr,&rcdtot,&ovfltot,
      &emptytot,&freeblk);
   printf("bldhsh: rcds %d "
      "ovfl %d empty %d "
      "free %d dups %d\n"
      "not found: gets %d updates %d\n",
      rcdtot, ovfltot, emptytot, freeblk,
      totdups, totnf, updtnf);
   /**********************************************************/
   hclose(hshhdr);
   return(0);
   } /* main */
